/*

Read bilinears, and combine into meson propagators 
(halving time axis by periodicity if 'DO_HALVE' is #define'd).

*/

#define LINT_ARGS 1      /* Enable Compiler Parameter Checking */
#include <stdio.h>

#define DO_HALVE

#define SPN_SIZE    4
#define COL_SIZE    3

#define NUM_OPERATORS 2     /* Here refers to different smearing regimes */

/* Spin "states" for bilinears ("diagonal" cases only)... */

typedef int SPIN_STATE ;

#define RANGE_SPIN_STATE 4 

#define P  0                /* pseudoscalar:                   G5 */
#define A  1                /* axial vector, 0th component: G0 G5 */
#define V  2                /* vector, 3rd component:          G3 */
#define T  3                /* tensor, 0-3 component:       G0 G3 */

#define p1 { 1,  0}
#define n1 {-1,  0}
/*
#define pi { 0,  1}
#define ni { 0, -1}
*/

static float spin_coeffs [RANGE_SPIN_STATE] [SPN_SIZE] [2] =
{
  {p1, p1, p1, p1},  /* P */
  {p1, p1, n1, n1},  /* A */
  {n1, p1, p1, n1},  /* V */
  {n1, p1, n1, p1}   /* T */
} ;


/* Meson spin combinations (for two ends of propagator)... */

typedef struct
{
  SPIN_STATE i, f;         /* initial, final                 */
  float      t_character   /* phase change under t-inversion */
} MES_PROP ;

#define NUM_MESONS  4

MES_PROP mes_combs [NUM_MESONS] =
{
  {P, P,  1.0},         /* pi  */
  {P, A, -1.0},         /* pi  */
  {A, A,  1.0},         /* pi  */
  {V, V,  1.0}          /* rho */
};


/* Array dimensions */

#define MAX_MASSES 16
#define MAX_LENGTH 48



int get_parameters (length, num_masses,
                    k_fermion_1, k_fermion_2)
int *length, *num_masses ;
float *k_fermion_1, *k_fermion_2 ;

/*

  Read run parameters from standard input

*/

{
  int i ;

  scanf ("%d %d", length, num_masses) ;
  if (*num_masses > MAX_MASSES)
  {
    fprintf (stderr, "ERROR:  Too many mass values\n") ;
    return (0) ;
  }
  else
  {
    for (i = 0; i < *num_masses; i++)
      scanf ("%f %f",  k_fermion_1 + i, k_fermion_2 + i) ; 
    return (1) ;
  }
}

int do_mesons (length, num_masses)
int length, num_masses ;
{
  if (length <= MAX_LENGTH)
  {
    typedef float VECTOR [MAX_LENGTH] [SPN_SIZE] [2] ;
    VECTOR bilins [MAX_MASSES] [NUM_OPERATORS] [SPN_SIZE] [COL_SIZE] ;
    int mass, t, s0, c0, st, o ;
    int meson ;

    /* Read in raw bilinear data */

    for (s0 = 0; s0 < SPN_SIZE; s0++)
      for (c0 = 0; c0 < COL_SIZE; c0++)
        for (mass = 0; mass < num_masses; mass++)
	  for (o = 0; o < NUM_OPERATORS; o++)
	    for (t = 0; t < length; t++)
	      for (st = 0; st < SPN_SIZE; st++)
	      {
		int result ;
		float *c_num = bilins [mass] [o] [s0] [c0] [t] [st] ;

		result = scanf ("%f %f", c_num, c_num + 1) ;
		if (result == EOF)
		{
		  if (s0 | c0 | mass | t | st)
		    printf ("\nERROR: Unexpected end of file\n") ;
		  return (0) ;
		}
	      }

    /* Accumulate meson propagators */

    for (mass = 0; mass < num_masses; mass++)
      for (meson = 0; meson < NUM_MESONS; meson++)
	{
	  MES_PROP *mes_spec = &(mes_combs [meson]) ;
	  float (*coeffs_i) [2] = spin_coeffs [mes_spec->i] ;
	  float (*coeffs_f) [2] = spin_coeffs [mes_spec->f] ;
	  float t_phase         = mes_spec->t_character ;
	
	  for (o = 0; o < NUM_OPERATORS; o++)
          {
#ifdef DO_HALVE
            float prop [(MAX_LENGTH / 2) + 1] [2] ;
#else
            float prop [MAX_LENGTH + 1] [2] ;
#endif
            int c ;

            for (t = 0; t < length; t++)
	    {
	      float re_sum = 0.0 ;
	      float im_sum = 0.0 ;

	      for (s0 = 0; s0 < SPN_SIZE; s0++)
		for (st = 0; st < SPN_SIZE; st++)
		{
		  float re_cc = (coeffs_i [s0] [0] * coeffs_f [st] [0]) -
		                (coeffs_i [s0] [1] * coeffs_f [st] [1]) ;
		  float im_cc = (coeffs_i [s0] [0] * coeffs_f [st] [1]) +
		                (coeffs_i [s0] [1] * coeffs_f [st] [0]) ;
		  float re_col_sum = 0.0 ;
		  float im_col_sum = 0.0 ;

		  for (c0 = 0; c0 < COL_SIZE; c0++)
		  {
		    re_col_sum += bilins [mass] [o] [s0] [c0] [t] [st] [0] ;
		    im_col_sum += bilins [mass] [o] [s0] [c0] [t] [st] [1] ;
		  }

		  re_sum += (re_cc * re_col_sum) - (im_cc * im_col_sum) ;
		  im_sum += (re_cc * im_col_sum) + (im_cc * re_col_sum) ;
		}
#ifdef DO_HALVE
	      if (t <= (length / 2))
	      {
		prop [t] [0]       = re_sum ;
		prop [t] [1]       = im_sum ;
	      }
	      else
	      {
		prop [length - t] [0] += t_phase * re_sum ;
		prop [length - t] [0] /= 2   ;
		prop [length - t] [1] += t_phase * im_sum ;
		prop [length - t] [1] /= 2   ;
	      }
#else
	      prop [t] [0]         = re_sum ;
	      prop [t] [1]         = im_sum ;
	      if (t == 0)
              {
		prop [length] [0]  = re_sum ;
		prop [length] [1]  = im_sum ;
	      }
#endif
	    }

	    /*
            for (c = 0; c < 2; c++)
	    */
            for (c = 0; c < 1; c++)
	    {
#ifdef DO_HALVE
	      for (t = 0; t < ((length / 2) + 1); t++)
#else
	      for (t = 0; t < (length + 1); t++)
#endif
		printf ("%13e\n", prop [t] [c]) ;

	      printf ("\n") ;
	    }
	  }
	  printf ("\n") ;
	}

    return 1 ;
  }
  else
  {
    fprintf (stderr, "ERROR: Lattice too long\n") ;
    return 0 ;
  }
}

int main (argc, argv, envp)
int argc ;
char *argv [] ;
char *envp [] ;
{
  int length, num_masses ;
  float k_fermion_1 [MAX_MASSES], k_fermion_2 [MAX_MASSES] ;

  if (get_parameters (&length, &num_masses, k_fermion_1, k_fermion_2))
  {
    int i ;

    printf ("\n") ;
#ifdef DO_HALVE
    printf ("%6d %6d %6d\n", num_masses,
	     NUM_MESONS * NUM_OPERATORS, (length / 2) + 1) ;
#else
    printf ("%6d %6d %6d\n", num_masses,
             NUM_MESONS * NUM_OPERATORS, length + 1) ;
#endif
    for (i = 0; i < num_masses; i++)
      printf ("%f %f\n", k_fermion_1 [i], k_fermion_2 [i]) ;
    printf ("\n") ;
    while (do_mesons (length, num_masses))
      printf ("\n") ;
  }
}

/*
--
--     End of File.
--
*/
